// #![allow(dead_code)]

mod cpu;
mod error;
mod instructions;
mod keyboard;
mod memory;
mod types;
mod video;

use crate::cpu::Cpu;
use crate::keyboard::Keyboard;
use crate::memory::Mem;
use crate::video::Crtc;

use clap::Parser;
use minifb::{Scale, ScaleMode, Window, WindowOptions};
use serde::{Deserialize, Serialize};

use std::fs::File;
use std::io::Read;
use std::process::exit;
use std::str::FromStr;
use std::sync::{mpsc, Mutex};
use std::{path::PathBuf, sync::Arc, thread};

use toml::Table;

//#[derive(Parser)]
//struct Cli {
//    // Load a rom onto the system rom
//    #[arg(short, required = true)]
//    rom: Option<std::path::PathBuf>,
//}

#[derive(Serialize, Deserialize, Debug)]
struct Config {
    char_rom: Option<String>,
    rom: String,
}

fn main() {
    let config: Config = match File::open("./config.toml") {
        Ok(mut file) => {
            let mut string = String::new();
            file.read_to_string(&mut string).unwrap();
            toml::from_str(string.as_str()).unwrap()
        }
        Err(_) => return,
    };
    println!("{config:#?}");
    let mut memory = Mem::new();
    let rom = match std::fs::File::open(config.rom) {
        Ok(file) => file,
        Err(error) => panic!("Couldn't open main rom! {:?}", error),
    };
    if let Err(error) = memory.load_rom(rom) {
        println!("{:?}", error);
    };

    let (interrupt_tx, interrupt_rx) = mpsc::channel();
    let (window_tx, window_rx) = mpsc::channel();

    memory
        .dump(PathBuf::from_str("./coredump.bin").unwrap())
        .unwrap();

    let shared_memory = Arc::new(Mutex::new(memory));
    let cpu_memory = shared_memory.clone();
    let display_memory = shared_memory.clone();
    let keyboard_memory = shared_memory.clone();

    thread::spawn(move || {
        let mut cpu = Cpu::new(cpu_memory, interrupt_rx);
        cpu.reset().unwrap();
        cpu.execute();
    });

    thread::spawn(move || {
        let mut screen = Crtc::new(
            display_memory,
            config.char_rom.as_ref(),
            interrupt_tx,
            window_tx,
        );
        screen.run();
    });

    let mut window = Window::new(
        "ʕ·ᴥ·ʔ-☆",
        512,
        380,
        WindowOptions {
            resize: true,
            borderless: true,
            title: true,
            transparency: false,
            scale: Scale::X2,
            scale_mode: ScaleMode::AspectRatioStretch,
            topmost: false,
            none: true,
        },
    )
    .unwrap();

    window.set_input_callback(Box::new(Keyboard::new(keyboard_memory)));

    while window.is_open() {
        let buffer = window_rx.recv().unwrap();
        window.update_with_buffer(&buffer, 512, 380).unwrap();
    }
}
